"""Compile-time values in Kirin IR.

This module contains the following:

## `abc` module
`abc.AttributeMeta`: The metaclass for all attributes.
`abc.Attribute`: The base class for all attributes.

## `types` module
`types.TypeAttribute`: The base class for all type attributes.
`types.PyClass`: A type attribute representing a Python class.
`types.TypeVar`: A type attribute representing a type variable.
`types.Literal`: A type attribute representing a literal type.
`types.Generic`: A type attribute representing a generic type.
`types.Union`: A type attribute representing a union type.

## `py` module

`py.PyAttr`: An attribute representing a Python value.
"""
